
from numpy import hstack, vstack, arange, newaxis, allclose
from bardeen.array_combi import reduce_arrays_align0, arrays_align0


def test_arrays_align0():
	basearr = hstack(arange(5)[:, newaxis] for k in range(5)) + vstack(arange(5) for k in range(5))
	arrays = [
		basearr[:5, :3],
		basearr[:3, :5],
		basearr[:4, :4],
	]
	for res in arrays_align0(arrays):
		assert res.shape == (3, 3)
		assert allclose(res, basearr[:3, :3])


def test_reduce_arrays_align0():
	basearr = hstack(arange(5)[:, newaxis] for k in range(5)) + vstack(arange(5) for k in range(5))
	arrays = [
		basearr[:5, :3],
		basearr[:3, :5],
		basearr[:4, :4],
	]
	res = reduce_arrays_align0(arrays)
	assert allclose(res, 3 * basearr[:3, :3])






"""
	unit tests for bardeen.storage
	tmpdir is automatically an empty temporary directory
	http://stackoverflow.com/questions/4199700/python-how-do-i-make-temporary-files-in-my-test-suite
"""

from os.path import join
from random import random
from bardeen.testutils import dict_round_floats
from bardeen.storage import store_array, load_array, store_array_bin, \
	load_array_bin, store_conf, load_conf, store_dict, load_dict
from numpy import array_equal, float64, exp, complex128
from numpy.random import rand, randint
from bardeen.testutils import get_random_word as rword, get_random_string as rstr


"""
	numpy arrays
"""


def store_load_array(tmpdir, original, header = '', load_dtype = float64):
	filename = join(unicode(tmpdir), '%s_%s.txt' % (unicode(original.dtype), 'x'.join(str(d) for d in original.shape)))
	store_array(original, filename, header = header)
	loaded = load_array(filename, dtype = load_dtype)
	assert array_equal(original, loaded)


def store_load_array_bin(tmpdir, original):
	filename = join(unicode(tmpdir), '%s_%s.npy' % (unicode(original.dtype), 'x'.join(str(d) for d in original.shape)))
	store_array_bin(original, filename)
	loaded = load_array_bin(filename)
	assert array_equal(original, loaded)


def test_float_array(tmpdir):
	original = rand(10, 15)
	store_load_array(tmpdir, original, header = 'random data columns')
	store_load_array_bin(tmpdir, original)


def test_int_array(tmpdir):
	original = randint(low = -4, high = 17, size = (10, 15))
	store_load_array(tmpdir, original, load_dtype = original.dtype)
	store_load_array_bin(tmpdir, original)


def test_complex_array(tmpdir):
	original = 100 * rand(10, 15) * exp(1j * rand(10, 15))
	store_load_array(tmpdir, original, header = '#complex numbers!!', load_dtype = complex128)
	store_load_array_bin(tmpdir, original)


"""
	dictionaries
"""

FLOAT_PREC = 8


def store_load_conf(tmpdir, original, header = ''):
	filename = join(unicode(tmpdir), 'test.conf')
	store_conf(original, filename, header = header)
	loaded = load_conf(filename)
	assert dict_round_floats(original, FLOAT_PREC) == dict_round_floats(loaded, FLOAT_PREC)


def store_load_dict(tmpdir, original):
	filename = join(unicode(tmpdir), 'test.conf')
	store_dict(original, filename)
	loaded = load_dict(filename)
	assert dict_round_floats(original, FLOAT_PREC) == dict_round_floats(loaded, FLOAT_PREC)


def store_load_dict_compressed(tmpdir, original):
	filename = join(unicode(tmpdir), 'test.conf')
	store_dict(original, filename, compressed = True)
	loaded = load_dict(filename, compressed = True)
	assert dict_round_floats(original, FLOAT_PREC) == dict_round_floats(loaded, FLOAT_PREC)


def test_mixed_dict(tmpdir):
	for N in range(50):
		dicti = {
			rstr(): rstr(),
			randint(-1e7, 1e7): randint(-1e7, 1e7),
			1e8 * random(): 1e8 * random(),
			rstr(): 1e8 * random(),
			randint(-1e7, 1e7): rstr(),
			1e8 * random(): randint(-1e7, 1e7),
			rstr(): None,
		}
		store_load_dict(tmpdir, dicti)
		store_load_dict_compressed(tmpdir, dicti)


def test_conf_dict(tmpdir):
	for N in range(50):
		dicti = {
			rword(): rstr(),
			randint(-1e7, 1e7): randint(-1e7, 1e7),
			1e8 * random(): 1e8 * random(),
			rword(): rstr(),
			rword(): randint(-1e7, 1e7),
			rword(): 1e8 * random(),
		}
		store_load_conf(tmpdir, dicti, header = '#test')
		dicti['hello world'] = 7
		try:
			store_load_conf(tmpdir, dicti, header = '#test')
		except KeyError:
			''' a key error should be thrown, as a space in a key would lead to inconsistent loading '''
		else:
			raise AssertionError('a key error should be thrown, as a space in a key would lead to inconsistent loading, but no error was thrown')
		del dicti['hello world']
		dicti['hello_world'] = '7\n'
		try:
			store_load_conf(tmpdir, dicti, header = '#test')
		except ValueError:
			''' a value error should be thrown, as a newline in a value would lead to inconsistent loading '''
		else:
			raise AssertionError('a value error should be thrown, as a newline in a value would lead to inconsistent loading, but no error was thrown')






from numpy import array


def arrays_align0(arrays):
	"""
	For a series of arrays, take the largest slice that fits within all arrays starting at 0.

	Number of dimensions for the arrays must be the same, but shape needn't be, if they can
	be aligned at the smallest index.
	"""
	if len(arrays) == 0:
		return []
	if len(arrays) == 1:
		return arrays
	shapes = tuple(arr.shape for arr in arrays)
	dim_counts = set(len(shape) for shape in shapes)
	assert len(dim_counts) == 1, ('number (not size) of dimensions should be the '
		'same for all arrays in reduce_arrays_align0; got lengths {0:}').format(
		''.join(str(dim) for dim in dim_counts))
	minshape = array(shapes).min(0)
	if any(dim == 0 for dim in minshape):
		return 0
	slices = tuple(slice(0, dim, 1) for dim in minshape)
	results = []
	for arr in arrays:
		results.append(arr[slices])
	return results


def reduce_arrays_align0(arrays, reduce_func=lambda a, b: a + b):
	"""
	For a series of arrays, take the largest slice that fits within all arrays starting at 0,
	and then apply a reduction function (on arr1 & arr2, then the result and arr3, etc).
	
	Number of dimensions for the arrays must be the same, but shape needn't be, if they can
	be aligned at the smallest index.
	
	If `reduce_func` is not replaced, the arrays are summed.
	"""
	arrays = arrays_align0(arrays)
	result = arrays[0]
	for arr in arrays[1:]:
		result = reduce_func(result, arr)
	return result






"""
	unit tests for bardeen.collection
"""

from bardeen.collection import group_by, without, make_unique_strs


def test_group_by():
	""" dividing numbers 0-100 in %N groups' """
	for N in [1, 2, 7, 12]:
		groups = group_by(range(70), lambda item: item % N)
		assert set(groups.keys()) == set(range(N))
		assert sum(len(items) for items in groups.values()) == 70
		for n, li in groups.items():
			assert all(k % N == n for k in li)
	""" grouping instances by class type """
	class A(object): pass
	class B(object): pass
	objs = [A() if (k % 2) * (k % 3) else B() for k in range(60)]
	groups = group_by(objs, lambda item: item.__class__)
	assert len(groups) == 2
	assert A in groups.keys() and B in groups.keys()
	assert len(groups[A]) == 20 and len(groups[B]) == 40
	assert all(item.__class__ is A for item in groups[A])


def test_without():
	li = list(range(5))
	assert list(without(li, (0, 2))) == [1, 3, 4]
	assert list(without(li, 3)) == [0, 1, 2, 4]
	print(without(li, 3))
	assert False


def test_make_unique_strs():
	assert make_unique_strs(['alpha', 'alpha', 'beta', 'alpha']) == ['alpha-1', 'alpha-2', 'beta', 'alpha-3']






"""
	tests for bardeen.testutils
"""

import string
from math import pi
from re import match
from bardeen.testutils import printable_line, get_random_string, get_random_word, create_file, dict_round_floats
from os.path import isfile, abspath
from numpy import float64


def check_letters_presence(word_generator, available_letters):
	"""
		check that the allowed characters actually occur; since the generation is random, there is always a chance that
		it didn't occur for the words finite amount of words generated; try N times such that the chance for any
		false positive is < 0.001 (0.1%)

		average letters generated: ``L = N * (16 + 2) / 2.``
		number of possible letters m: ``m = len(available_letters)``
		probability of a specific letter to not occur p: ``p = ((m - 1.) / m) ** L``
		therefore, minimum N such that m * p < 0.001:
		((m - 1.) / m) ** L = 0.001 / m
		L * ln ((m - 1.) / m) = ln (0.001 / m)
		N = (2. / 18) * ln (0.001 / m) / ln ((m - 1.) / m)
		which for m 10 up to 100k is m < N < ~2m
	"""
	for check in available_letters:
		for k in range(2 * len(available_letters)):
		#for k in range(1):
			if check in word_generator():
				break
		else:
			m = len(available_letters)
			prob_false_neg = m * (((m - 1.) / m) ** (18 * m))
			raise AssertionError('didn\'t find "%s" in %d words (could be random, but p = %.6f)' % (check, 2 * m, prob_false_neg))


def test_get_random_word():
	"""
		test if get_random_word is within the limits and generated the allowed characters
	"""
	for k in range(50):
		word = get_random_word()
		assert match(r'[\w\d]{2,16}', word)
	check_letters_presence(get_random_word, string.letters + string.digits)


def test_get_random_string():
	"""
		test if get_random_string is within the limits and generated the allowed characters
	"""
	for k in range(50):
		word = get_random_string()
		assert match(r'[\w\d\s!"#$%&\'\(\)\*\+,\-\./:;<=>\?@\[\\\]^_`\{|\}~]{2,16}', word)
	check_letters_presence(get_random_string, printable_line)


def test_create_file(tmpdir):
	file_path = create_file(unicode(tmpdir))
	assert isfile(file_path)
	assert abspath(file_path).startswith(abspath(unicode(tmpdir))), 'check correct directory'


def test_dict_round_floats():
	test_dict = {
		pi: float64(pi),
		'3.1415': None,
	}
	validate_dict = {
		3.14: float64(3.14),
		'3.1415': None,
	}
	dict_round_floats(test_dict, 2)
	assert test_dict == validate_dict






"""
	unit tests for bardeen.inout
"""

from time import sleep
from bardeen.inout import clear_stdout, reprint, add_linebreaks

"""
	clear_stdout and reprint and difficult to test automatically;
	optionally you can run the below code manually in a shell
"""


def manual_test_clear_stdout():
	"""
		Prints "before going back", then creates lines, removes them and prints numbers
	"""
	for k in range(5):
		print('this line should be removed; if you see it, it went wrong')
	clear_stdout(lines = 5)
	for k in range(5):
		print('item #%d' % (k + 1))


def manual_test_reprint():
	"""
		Simple status monitoring example, showing count and stars (takes 3 lines but uses 2)
	"""
	print('\n' * 2)
	for k in range(1, 11):
		reprint('item #%.2d/10: %s \n^ that is the status!' % (k, '*' * k), lines = 3)
		sleep(0.15)


def test_add_linebreaks():
	assert add_linebreaks('hello world test me please', max_len=8) == \
		'hello\nworld\ntest me\nplease'
	assert add_linebreaks('test test superlongwordthatdoesn\'tfit test test', max_len=10) == \
		'test test\nsuperlongwordthatdoesn\'tfit\ntest test'
	assert add_linebreaks('multiple     spaces   yet more spaces        here', max_len=20) == \
		'multiple     spaces\n  yet more spaces   \n    here'
	# not sure this is right, first one seems to be after 19...
	assert add_linebreaks('', max_len=8) == ''


if __name__ == '__main__':
	print('manual_test_clear_stdout: you should see item#1-5')
	manual_test_clear_stdout()
	print('manual_test_reprint: you should see an updating status')
	manual_test_reprint()






from pytest import raises
from numpy import prod
from grid import to_coordinate, from_coordinate


def test_grid_inversion():
	dims = (2, 3, 2, 4)
	for k in range(prod(dims)):
		assert from_coordinate(to_coordinate(k, dims), dims) == k


def test_grid_examples():
	dims = (2, 3, 2, 4)
	assert to_coordinate(0, dims)  == (0, 0, 0, 0)
	assert to_coordinate(17, dims) == (0, 2, 0, 1)
	assert to_coordinate(39, dims) == (1, 1, 1, 3)


def test_changing_dims():
	dims = (2, 3, 2, 4)
	assert to_coordinate(17, dims) == (0, 2, 0, 1)
	dims = (1, 2, 1, 1)
	assert to_coordinate(1, dims) == (0, 1, 0, 0)


def test_one_dimensional():
	dims = (7,)
	for k in range(dims[0]):
		assert to_coordinate(k, dims)  == (k,)
		assert from_coordinate(to_coordinate(k, dims), dims) == k


def test_out_of_bounds():
	dims = (2, 3, 2, 4)
	with raises(AssertionError):
		to_coordinate(-1, dims)
	with raises(AssertionError):
		print(to_coordinate(prod(dims), dims))
	with raises(AssertionError):
		from_coordinate((0, 0, 0, -1), dims)
	with raises(AssertionError):
		from_coordinate((0, 0, 0, 7), dims)






"""
	unit tests for bardeen.system
"""

from os.path import join, isdir, isfile, islink
from bardeen.system import mkdirp, link_else_copy
from bardeen.testutils import create_file, get_random_word
from shutil import Error


def test_link_else_copy(tmpdir):
	old_path = create_file(unicode(tmpdir))
	new_path = join(unicode(tmpdir), get_random_word()) + '.txt'
	link_else_copy(old_path, new_path)
	assert isfile(new_path) or islink(new_path)
	assert open(old_path, 'r').read() == open(new_path, 'r').read(), 'check if files the same'
	try:
		link_else_copy(old_path, new_path)
	except Error:
		""" link already created """
	except OSError:
		""" file already copied """
	else:
		raise AssertionError('link_else_copy does not raise an error if the destination file already exists')
	try:
		link_else_copy(old_path + '.DoesNotExist', new_path)
	except IOError:
		""" as expected """
	else:
		raise AssertionError('link_else_copy does not raise an error if the source file doesn\'t exist')


def test_mkdirp(tmpdir):
	dir_pth = join(unicode(tmpdir), get_random_word(), get_random_word(), get_random_word())
	mkdirp(dir_pth)
	assert isdir(dir_pth)
	file_pth = join(dir_pth, get_random_word())
	fh = open(file_pth, 'w+')
	fh.write('testfile')
	fh.close()
	assert isfile(file_pth)
	try:
		mkdirp(file_pth)
	except OSError:
		""" expected error """
	else:
		raise AssertionError('did not throw an exception, even though the path was a file and as such, \
			no directory was created')






"""
	unit tests for bardeen.numerical
"""

from bardeen.numerical import laplacian


def test_laplacian():
	for N in [1, 2, 3, 7]:
		M = laplacian(N)
		assert M.shape == (N, N)
		for k in range(N):
			for m in range(N):
				if k == m:
					assert M[k, m] == 2
				elif abs(k - m) == 1:
					assert M[k, m] == -1
				else:
					assert M[k, m] == 0


